import importlib.resources

import numpy as np
from PyQt5 import QtCore, QtWidgets, uic
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QKeySequence
from PyQt5.QtWidgets import QShortcut

from .. import scores


class CheckableComboBox(QtWidgets.QComboBox):
    def __init__(self, *args, **kwargs):
        super(CheckableComboBox, self).__init__(*args, **kwargs)
        self._changed = False
        self.view().pressed.connect(self.handleItemPressed)

    def addItem(self, text, userData):
        super(CheckableComboBox, self).addItem(text, userData)
        item = self.model().item(self.count() - 1, self.modelColumn())
        item.setCheckState(Qt.Unchecked)

    def handleItemPressed(self, index):
        item = self.model().itemFromIndex(index)
        checked = item.checkState() == Qt.Checked
        item.setCheckState(Qt.Unchecked if checked else Qt.Checked)
        self._changed = True

    def hidePopup(self):
        if not self._changed:
            super(CheckableComboBox, self).hidePopup()
        self._changed = False

    def itemChecked(self, index):
        item = self.model().item(index, self.modelColumn())
        return item.checkState() == Qt.Checked

    def itemsCheckedData(self):
        data_checked = []
        for ii in range(self.count()):
            item = self.model().item(ii, self.modelColumn())
            if item.checkState() == Qt.Checked:
                # get the actual item
                data_checked.append(self.itemData(ii))
        return data_checked

    def setItemChecked(self, index, checked=False):
        item = self.model().item(index, self.modelColumn())
        item.setCheckState(Qt.Checked if checked else Qt.Unchecked)


class LabelButtonWidget(QtWidgets.QWidget):
    button_pressed = QtCore.pyqtSignal(str)

    def __init__(self, feature, *args, **kwargs):
        super(LabelButtonWidget, self).__init__(*args, **kwargs)
        self.feature = feature
        self.shortcut = None

        self.verticalLayout = QtWidgets.QVBoxLayout(self)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.pushButton = QtWidgets.QPushButton("label")
        self.pushButton.setToolTip(scores.get_feature_label(feature))
        self.pushButton.pressed.connect(self.on_button)
        self.verticalLayout.addWidget(self.pushButton)
        self.comboBox = QtWidgets.QComboBox(self)
        shortcut = scores.get_feature_shortcut(feature)
        self.comboBox.addItem(shortcut)
        for sc in "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ":
            self.comboBox.addItem(sc)
        self.comboBox.currentTextChanged.connect(self.on_combobox)
        self.on_combobox(shortcut)  # initial shortcut
        self.verticalLayout.addWidget(self.comboBox)
        # Initialize button label
        self.set_score(np.nan)

    @QtCore.pyqtSlot()
    def on_button(self):
        self.button_pressed.emit(self.feature)

    @QtCore.pyqtSlot(str)
    def on_combobox(self, shortcut=None):
        if shortcut is None:
            shortcut = self.comboBox.currentText()
        seq = QKeySequence(shortcut)
        self.pushButton.setShortcut(seq)

    def set_score(self, score=np.nan):
        """Add square brackets in the button"""
        label = self.feature[-3:].upper()
        if score is True:
            label = f"[{label}]"
        elif score is False:
            label = f"!{label}"
        self.pushButton.setText(label)
        # Somehow necessary to re-enable the shortcut
        self.on_combobox(None)


class TabMultiClassLabel(QtWidgets.QWidget):
    """Tab for doing binary classification"""

    def __init__(self, *args, **kwargs):
        super(TabMultiClassLabel, self).__init__(*args, **kwargs)

        ref = importlib.resources.files("dctag.gui") / "tab_multiple.ui"
        with importlib.resources.as_file(ref) as path_ui:
            uic.loadUi(path_ui, self)

        self.session = None
        self.event_index = 0

        self.settings = QtCore.QSettings()

        # populate ML scores combobox
        self.comboBox_score.clear()
        for feat in scores.get_dctag_label_dict(
                name=self.settings.value("labeling group", "ml_scores_blood")):
            flabel = f"{scores.get_feature_label(feat)} [{feat[-3:].upper()}]"
            self.comboBox_score.addItem(flabel, feat)

        # signals
        self.pushButton_start.clicked.connect(self.on_start)
        self.pushButton_next.clicked.connect(self.on_event_button)
        self.pushButton_prev.clicked.connect(self.on_event_button)
        self.pushButton_fast_next.clicked.connect(self.on_event_button)
        self.pushButton_fast_prev.clicked.connect(self.on_event_button)
        self.toolButton_reset.clicked.connect(self.on_event_button)
        self.spinBox_jump_to.valueChanged.connect(self.on_jump_to)

        self.toolButton_reset.setIcon(self.style().standardIcon(
            QtWidgets.QStyle.SP_TrashIcon))

        # keyboard shortcuts
        self.shortcuts = []
        for button, shortcuts in [
            [self.pushButton_next, ["Right"]],
            [self.pushButton_prev, ["Left"]],
            [self.pushButton_fast_prev, ["Shift+Left"]],
            [self.pushButton_fast_next, ["Shift+Right"]],
        ]:
            for seq in shortcuts:
                sc = QShortcut(QKeySequence(seq), self)
                sc.activated.connect(button.click)
                # include original ToolTip
                tt = button.toolTip()
                tt = tt + "; " if tt else ""
                button.setToolTip(f"{tt}Shortcuts: {', '.join(shortcuts)}")
                self.shortcuts.append(sc)  # keep a reference

        #: list of buttons for labeling
        self.label_buttons = []

    @property
    def features(self):
        datas = self.comboBox_score.itemsCheckedData()
        return datas

    def update_session(self, session):
        """Update this widget with the session info"""
        # Whenever the user leaves and comes back to this tab, he has
        # to lock-in again to label data.
        self.lock_out()
        if self.session is not session:
            self.session = session
            self.event_index = 0
        if self.session:
            self.setEnabled(True)
            self.spinBox_jump_to.setMaximum(self.session.event_count)
            self.goto_event(self.event_index)
        else:
            self.setEnabled(False)
            self.widget_vis.reset(reset_plots=True)

    def goto_event(self, index):
        if index < 0:
            self.goto_event(0)
            return
        elif index >= self.session.event_count:
            self.goto_event(self.session.event_count - 1)
            return

        self.event_index = index

        # enable/disable skip buttons
        self.pushButton_prev.setDisabled(index == 0)
        self.pushButton_next.setDisabled(index == self.session.event_count - 1)

        # handle previous and next score labels
        if index != 0 and self.features:
            candidates = self.session.get_scores_true(index - 1)
            scs = [f for f in candidates if f in self.session.linked_features]
            label = " ".join([s[-3:] for s in scs]).upper() or "nan"
            self.label_score_prev.setText(label)
        else:
            self.label_score_prev.setText("")

        if index != self.session.event_count - 1 and self.features:
            candidates = self.session.get_scores_true(index + 1)
            scs = [f for f in candidates if f in self.session.linked_features]
            label = " ".join([s[-3:] for s in scs]).upper() or "nan"
            self.label_score_next.setText(label)
        else:
            self.label_score_next.setText("")

        # indicate current score label
        if self.features:
            for button in self.label_buttons:
                score = self.session.get_score(button.feature, index)
                button.set_score(score)

        # update spinBox_jump_to
        self.spinBox_jump_to.blockSignals(True)
        self.spinBox_jump_to.setValue(self.event_index + 1)
        self.spinBox_jump_to.blockSignals(False)

        # update progress bar
        if self.features:
            fscores = self.session.scores_cache.get(self.features[0], [])
            num_rated = np.sum(~np.isnan(fscores))
            perc = int(np.floor(num_rated / self.session.event_count * 100))
            self.progressBar.setValue(perc)

        # visualization
        self.widget_vis.set_event(self.session, index)

    def lock_in(self):
        """Begin labeling"""
        self.pushButton_start.setVisible(False)
        self.comboBox_score.setEnabled(False)
        self.progressBar.setVisible(True)
        self.widget_label_keys.setEnabled(True)
        main = QtWidgets.QApplication.activeWindow()
        lids = [scores.get_feature_label(ft)[-3:] for ft in self.features]
        main.set_title("-".join(lids).upper())
        # clear current layout of label buttons
        for widgetrm in self.label_buttons:
            widgetrm.deleteLater()
            widgetrm.setParent(None)
        self.label_buttons.clear()
        # populate the score buttons and connect signals
        for feat in self.features:
            fbutton = LabelButtonWidget(feat)
            fbutton.button_pressed.connect(self.on_event_button_feature)
            self.layout_label_buttons.addWidget(fbutton)
            self.label_buttons.append(fbutton)

    def lock_out(self):
        """Stop labeling"""
        self.pushButton_start.setVisible(True)
        self.comboBox_score.setEnabled(True)
        self.progressBar.setVisible(False)
        self.widget_label_keys.setEnabled(False)

    @QtCore.pyqtSlot()
    def on_event_button(self):
        btn = self.sender()
        if btn is self.pushButton_next:
            self.goto_event(self.event_index + 1)
        elif btn is self.pushButton_prev:
            self.goto_event(self.event_index - 1)
        elif btn is self.pushButton_fast_prev:
            for ii in range(1, self.event_index):
                new_index = self.event_index - ii
                curscores = self.session.get_scores_true(new_index)
                if set(self.features).isdisjoint(set(curscores)):
                    break
            else:
                new_index = 0
            self.goto_event(new_index)
        elif btn is self.pushButton_fast_next:
            start = min(self.event_index + 1, self.session.event_count - 1)
            for new_index in range(start, self.session.event_count):
                curscores = self.session.get_scores_true(new_index)
                if set(self.features).isdisjoint(set(curscores)):
                    break
            else:
                new_index = self.session.event_count - 1
            self.goto_event(new_index)
        elif btn is self.toolButton_reset:
            # linked features will also be reset
            self.session.reset_score(self.features[0], self.event_index)
            self.goto_event(self.event_index + 1)

    @QtCore.pyqtSlot(str)
    def on_event_button_feature(self, feature):
        self.session.set_score(feature, self.event_index, True)
        self.goto_event(self.event_index + 1)

    @QtCore.pyqtSlot(int)
    def on_jump_to(self, event_index):
        self.goto_event(event_index - 1)

    @QtCore.pyqtSlot()
    def on_start(self):
        if not self.features:
            QtWidgets.QMessageBox.warning(
                self,
                "No score features chosen.",
                "Please select ML score features in the combobox first!"
                )
        else:
            self.session.linked_features = self.features
            self.session.autocomplete_linked_features()
            self.lock_in()
            self.goto_event(0)



import pathlib

from PyQt5 import QtCore, QtWidgets
import pytest

from dctag import session
from .helper import get_clean_data_path


data_dir = pathlib.Path(__file__).parent / "data"


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 3000)
    pass
    # A test function will be run at this point
    yield
    # Code that will run after your test
    # restore dctag-tester for other tests
    QtCore.QCoreApplication.setOrganizationName("MPL")
    QtCore.QCoreApplication.setOrganizationDomain("dc-cosmos.org")
    QtCore.QCoreApplication.setApplicationName("dctag")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.setValue("user/name", "dctag-tester")
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 3000)


def test_error_session(qtbot, mw):
    path = get_clean_data_path()
    with session.DCTagSession(path, "dctag-tester") as dts:
        dts.set_score("ml_score_r1f", 0, True)
    # open session
    mw.on_action_open(path)
    # got to other tab and remove the file
    mw.tabWidget.setCurrentIndex(1)
    assert not mw.tab_session.plainTextEdit_logs.toPlainText() == "No session."
    path.unlink()
    mw.tabWidget.setCurrentIndex(0)
    assert mw.tab_session.plainTextEdit_logs.toPlainText().startswith(
        "Cannot get logs from")


def test_view_session(qtbot, mw):
    """Clearing the session should not cause any trouble"""
    path = get_clean_data_path()
    # make sure there is no session
    assert mw.tab_session.plainTextEdit_logs.toPlainText() == "No session."
    # claim session
    with session.DCTagSession(path, "dctag-tester") as dts:
        dts.set_score("ml_score_r1f", 0, True)
        dts.set_score("ml_score_r1f", 1, False)
        dts.set_score("ml_score_r1f", 2, True)
        dts.set_score("ml_score_r1f", 3, False)
    # open session
    mw.on_action_open(path)
    assert mw.tab_session.plainTextEdit_logs.toPlainText().count(
        "ml_score_r1f")
    mw.on_action_close()
    assert mw.tab_session.plainTextEdit_logs.toPlainText() == "No session."



import atexit
import shutil
import tempfile
import time

from dctag.gui.main import DCTag

import pytest
from PyQt5 import QtCore, QtWidgets

TMPDIR = tempfile.mkdtemp(prefix=time.strftime(
    "dctag_test_%H.%M_"))


@pytest.fixture
def mw(qtbot):
    # Always set server correctly, because there is a test that
    # makes sure DCOR-Aid starts with a wrong server.
    QtWidgets.QApplication.processEvents(
        QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 200)
    # Code that will run before your test
    mw = DCTag()
    qtbot.addWidget(mw)
    QtWidgets.QApplication.setActiveWindow(mw)
    QtWidgets.QApplication.processEvents(
        QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 200)
    # Run test
    yield mw
    # Make sure that all daemons are gone
    mw.close()
    # It is extremely weird, but this seems to be important to avoid segfaults!
    QtWidgets.QApplication.processEvents(
        QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 200)


def pytest_configure(config):
    """
    Allows plugins and conftest files to perform initial configuration.
    This hook is called for every plugin and initial conftest
    file after command line options have been parsed.
    """
    tempfile.tempdir = TMPDIR

    # Default settings
    QtCore.QCoreApplication.setOrganizationName("MPL")
    QtCore.QCoreApplication.setOrganizationDomain("dc-cosmos.org")
    QtCore.QCoreApplication.setApplicationName("dctag")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.setValue("user/name", "dctag-tester")
    settings.setValue("debug/without timers", "1")
    settings.setValue("labeling group", "ml_scores_blood")
    atexit.register(shutil.rmtree, TMPDIR, ignore_errors=True)


def pytest_unconfigure(config):
    """
    called before test process is exited.
    """
    QtCore.QCoreApplication.setOrganizationName("MPL")
    QtCore.QCoreApplication.setOrganizationDomain("dc-cosmos.org")
    QtCore.QCoreApplication.setApplicationName("dctag")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.remove("user/name")
    settings.remove("debug/without timers")



import pyqtgraph as pg


class SimpleImageView(pg.ImageView):
    """Custom class for visualizing images"""

    def __init__(self, *args, **kwargs):
        super(SimpleImageView, self).__init__(*args, **kwargs)

        # disable pyqtgraph controls we don't need
        self.ui.histogram.hide()
        self.ui.roiBtn.hide()
        self.ui.menuBtn.hide()
        # disable keyboard shortcuts
        self.keyPressEvent = lambda _: None
        self.keyReleaseEvent = lambda _: None

    def getViewBox(self):
        return self.getPlotItem().getViewBox()



import importlib.resources

import numpy as np
from PyQt5 import QtCore, QtWidgets, uic
from PyQt5.QtGui import QKeySequence
from PyQt5.QtWidgets import QShortcut

from .. import scores


class TabBinaryLabel(QtWidgets.QWidget):
    """Tab for doing binary classification"""

    def __init__(self, *args, **kwargs):
        super(TabBinaryLabel, self).__init__(*args, **kwargs)

        ref = importlib.resources.files("dctag.gui") / "tab_binary.ui"
        with importlib.resources.as_file(ref) as path_ui:
            uic.loadUi(path_ui, self)

        self.session = None
        self.event_index = 0

        # settings
        self.settings = QtCore.QSettings()

        # populate ML scores combobox
        self.comboBox_score.clear()
        for feat in scores.get_dctag_label_dict(
                name=self.settings.value("labeling group", "ml_scores_blood")):
            flabel = f"{scores.get_feature_label(feat)} [{feat[-3:].upper()}]"
            self.comboBox_score.addItem(flabel, feat)

        # signals
        self.pushButton_start.clicked.connect(self.on_start)
        self.pushButton_next.clicked.connect(self.on_event_button)
        self.pushButton_prev.clicked.connect(self.on_event_button)
        self.pushButton_yes.clicked.connect(self.on_event_button)
        self.pushButton_no.clicked.connect(self.on_event_button)
        self.pushButton_fast_next.clicked.connect(self.on_event_button)
        self.pushButton_fast_prev.clicked.connect(self.on_event_button)
        self.toolButton_reset.clicked.connect(self.on_event_button)
        self.spinBox_jump_to.valueChanged.connect(self.on_jump_to)

        self.toolButton_reset.setIcon(self.style().standardIcon(
            QtWidgets.QStyle.SP_TrashIcon))
        self.toolButton_reset.repaint()

        # keyboard shortcuts
        self.shortcuts = []
        for button, shortcuts in [
            [self.pushButton_yes, ["Up", "J", "Y"]],
            [self.pushButton_no, ["Down", "F", "N"]],
            [self.pushButton_next, ["Right"]],
            [self.pushButton_prev, ["Left"]],
            [self.pushButton_fast_prev, ["Shift+Left"]],
            [self.pushButton_fast_next, ["Shift+Right"]],
        ]:
            for seq in shortcuts:
                sc = QShortcut(QKeySequence(seq), self)
                sc.activated.connect(button.click)
                self.shortcuts.append(sc)  # keep a reference
            # append shortcuts tool tip to original tool tip
            tt = button.toolTip()
            tt = tt + "; " if tt else ""
            button.setToolTip(f"{tt}Shortcuts: {', '.join(shortcuts)}")

    @property
    def feature(self):
        return self.comboBox_score.currentData()

    def update_session(self, session):
        """Update this widget with the session info"""
        # Whenever the user leaves and comes back to this tab, he has
        # to lock-in again to label data.
        self.lock_out()
        if self.session is not session:
            self.session = session
            self.event_index = 0
        if self.session:
            self.setEnabled(True)
            self.spinBox_jump_to.setMaximum(self.session.event_count)
            self.goto_event(self.event_index)
        else:
            self.setEnabled(False)
            self.widget_vis.reset(reset_plots=True)

    def goto_event(self, index):
        if index < 0:
            self.goto_event(0)
            return
        elif index >= self.session.event_count:
            self.goto_event(self.session.event_count - 1)
            return

        self.event_index = index

        # enable/disable skip buttons
        self.pushButton_prev.setDisabled(index == 0)
        self.pushButton_next.setDisabled(index == self.session.event_count - 1)

        # handle previous and next score labels
        if index != 0 and self.feature:
            prev_score = self.session.get_score(self.feature, index - 1)
            if not np.isnan(prev_score):
                prev_score = "Yes" if prev_score else "No"
            self.label_score_prev.setText(f"{prev_score}")
        else:
            self.label_score_prev.setText("")

        if index != self.session.event_count - 1 and self.feature:
            next_score = self.session.get_score(self.feature, index + 1)
            if not np.isnan(next_score):
                next_score = "Yes" if next_score else "No"
            self.label_score_next.setText(f"{next_score}")
        else:
            self.label_score_next.setText("")

        # indicate current score label
        yes = "Yes"
        no = "No"
        if self.feature:
            current_score = self.session.get_score(self.feature, index)
            if not np.isnan(current_score):
                if current_score:
                    yes = "[Yes]"
                else:
                    no = "[No]"
        self.pushButton_no.setText(no)
        self.pushButton_yes.setText(yes)

        # update spinBox_jump_to
        self.spinBox_jump_to.blockSignals(True)
        self.spinBox_jump_to.setValue(self.event_index + 1)
        self.spinBox_jump_to.blockSignals(False)

        # update progress bar
        if self.feature:
            fscores = self.session.scores_cache.get(self.feature, [])
            num_rated = np.sum(~np.isnan(fscores))
            perc = int(np.floor(num_rated / self.session.event_count * 100))
            self.progressBar.setValue(perc)

        # visualization
        self.widget_vis.set_event(self.session, index)

    def lock_in(self):
        """Begin labeling"""
        self.pushButton_start.setVisible(False)
        self.comboBox_score.setEnabled(False)
        self.progressBar.setVisible(True)
        self.widget_label_keys.setEnabled(True)
        main = QtWidgets.QApplication.activeWindow()
        assert self.feature is not None, "feature not set!"
        label = scores.get_feature_label(self.feature)
        main.set_title(f"{self.feature[-3:].upper()}: {label}")

    def lock_out(self):
        """Stop labeling"""
        self.pushButton_start.setVisible(True)
        self.comboBox_score.setEnabled(True)
        self.progressBar.setVisible(False)
        self.widget_label_keys.setEnabled(False)

    @QtCore.pyqtSlot()
    def on_event_button(self):
        btn = self.sender()
        if btn is self.pushButton_next:
            self.goto_event(self.event_index + 1)
        elif btn is self.pushButton_prev:
            self.goto_event(self.event_index - 1)
        elif btn is self.pushButton_no:
            self.session.set_score(self.feature, self.event_index, False)
            self.goto_event(self.event_index + 1)
        elif btn is self.pushButton_yes:
            self.session.set_score(self.feature, self.event_index, True)
            self.goto_event(self.event_index + 1)
        elif btn is self.pushButton_fast_prev:
            for ii in range(1, self.event_index):
                new_index = self.event_index - ii
                if np.isnan(self.session.get_score(self.feature, new_index)):
                    break
            else:
                new_index = 0
            self.goto_event(new_index)
        elif btn is self.pushButton_fast_next:
            start = min(self.event_index + 1, self.session.event_count - 1)
            for new_index in range(start, self.session.event_count):
                if np.isnan(self.session.get_score(self.feature, new_index)):
                    break
            else:
                new_index = self.session.event_count - 1
            self.goto_event(new_index)
        elif btn is self.toolButton_reset:
            self.session.reset_score(self.feature, self.event_index)
            self.goto_event(self.event_index + 1)

    @QtCore.pyqtSlot(int)
    def on_jump_to(self, event_index):
        self.goto_event(event_index - 1)

    @QtCore.pyqtSlot()
    def on_start(self):
        self.session.linked_features = []
        self.lock_in()
        self.goto_event(0)



import getpass
import importlib.resources
import pathlib
import signal
import sys
import time
import traceback

import dclab
import h5py
import numpy
from PyQt5 import uic, QtCore, QtWidgets
import pyqtgraph as pg

from .. import scores
from .. import session
from .._version import version


# global plotting configuration parameters
pg.setConfigOption("background", None)
pg.setConfigOption("foreground", "k")
pg.setConfigOption("antialias", True)
pg.setConfigOption("imageAxisOrder", "row-major")


class DCTag(QtWidgets.QMainWindow):
    def __init__(self):
        super(DCTag, self).__init__()

        # Settings are stored in the .ini file format. Even though
        # `self.settings` may return integer/bool in the same session,
        # in the next session, it will reliably return strings. Lists
        # of strings (comma-separated) work nicely though.
        # Some promoted widgets need the below constants set in order
        # to access the settings upon initialization.
        QtCore.QCoreApplication.setOrganizationName("MPL")
        QtCore.QCoreApplication.setOrganizationDomain("dc-cosmos.org")
        QtCore.QCoreApplication.setApplicationName("dctag")
        QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)

        # if "--version" was specified, print the version and exit
        if "--version" in sys.argv:
            print(version)
            QtWidgets.QApplication.processEvents(
                QtCore.QEventLoop.AllEvents, 300)
            sys.exit(0)

        #: DCOR-Aid settings
        self.settings = QtCore.QSettings()
        username = self.settings.value("user/name", None)
        if username is None:
            username, ok = QtWidgets.QInputDialog.getText(
                self,
                "Specify username",
                "Please specify your alias. This will be used to<br>"
                + "attribute the datasets you labeled to a unique<br>"
                + "entity and it will be used to lock/link .rtdc files<br>"
                + "to that entity.<br><br>"
                + "Please choose wisely.<br>"
                + "You should not change it in the future.",
                text=getpass.getuser()
            )
            if ok and username.strip():
                username = username.strip()
            else:
                # Abort
                sys.exit(0)
        self.settings.setValue("user/name", username)

        # initialize UI
        ref = importlib.resources.files("dctag.gui") / "main.ui"
        with importlib.resources.as_file(ref) as path_ui:
            uic.loadUi(path_ui, self)

        self.set_title()
        # Disable native menubar (e.g. on Mac)
        self.menubar.setNativeMenuBar(False)
        # File menu
        self.actionOpen.triggered.connect(self.on_action_open)
        self.actionQuit.triggered.connect(self.on_action_quit)
        self.actionClose.triggered.connect(self.on_action_close)
        # Preferences menu
        self.actionSelectLabels.triggered.connect(self.on_action_select_labels)
        # Session menu
        self.actionFlushSession.triggered.connect(self.on_action_flush)
        self.actionBackupSession.triggered.connect(self.on_action_backup)
        # Help menu
        self.actionSoftware.triggered.connect(self.on_action_software)
        self.actionAbout.triggered.connect(self.on_action_about)

        # tabwidget
        self.tabWidget.currentChanged.connect(self.on_tab_changed)

        #: holds the current DCTagSession instance
        self.session = None

        self.show()
        self.raise_()
        self.activateWindow()

        # flush session regularly
        if bool(int(self.settings.value("debug/without timers", "0"))):
            self.timer = None
        else:
            self.timer = QtCore.QTimer()
            self.timer.timeout.connect(self.session_flush_statusbar)
            self.timer.start(60000)

    def closeEvent(self, event):
        if self.session_close():
            event.accept()
        else:
            event.ignore()

    def dragEnterEvent(self, e):
        """Whether files are accepted"""
        if e.mimeData().hasUrls():
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, e):
        """Add dropped files to view"""
        urls = e.mimeData().urls()
        if len(urls) > 1:
            raise ValueError(
                f"Can only open one file at a time, got {len(urls)}!")
        elif urls:
            pp = pathlib.Path(urls[0].toLocalFile())
            self.session_open(pp)

    @QtCore.pyqtSlot()
    def on_action_about(self):
        about_text = "DCTag: Annotate .rtdc files for ML training"
        QtWidgets.QMessageBox.about(self, f"DCTag {version}", about_text)

    @QtCore.pyqtSlot()
    def on_action_backup(self):
        """Create an .rtdc file with the current history and score cache"""
        reply = QtWidgets.QMessageBox.question(
            self,
            "Do you know what you are doing?",
            "You are about to export your ML scores to an HDF5 file. "
            + "Normally, you do not need to do this. You may want to, if you "
            + "are not able to flush the current session (e.g. because you "
            + "have your .rtdc file on a network share and there is no "
            + "connectivity). Proceed?"
        )
        if reply == QtWidgets.QMessageBox.Yes:
            # First attempt to flush the session again.
            try:
                self.session.flush()
            except BaseException:
                pass  # never mind
            # Open a dialog for the user to save the scores as an .hdf5 file
            po, _ = QtWidgets.QFileDialog.getSaveFileName(
                self, 'Path to new ML score file', '', 'HDF5 file (*.h5)')
            if po:
                po = pathlib.Path(po)
                # make sure the suffix is .h5
                if po.suffix != ".h5":
                    po = po.with_name(po.name + ".h5")
                self.session.backup_scores(po)
            # Ask the user whether to close DCTag now
            reply2 = QtWidgets.QMessageBox.question(
                self,
                "Close DCTag now?",
                "You have saved your scores and somebody with sufficient "
                + "experience will be able to append them to the original "
                + "file (if that still exists). Close DCTag now?")
            if reply2 == QtWidgets.QMessageBox.Yes:
                self.on_action_quit(force=True)

    @QtCore.pyqtSlot()
    def on_action_close(self):
        self.session_close()
        # Go to session tab and update info
        self.tabWidget.setCurrentIndex(0)
        self.on_tab_changed()

    @QtCore.pyqtSlot()
    def on_action_flush(self):
        if self.session:
            self.session.flush()

    @QtCore.pyqtSlot()
    def on_action_open(self, path=None):
        if path is None:
            path, _ = QtWidgets.QFileDialog.getOpenFileName(
                self,
                'Select RT-DC data',
                self.settings.value("paths/open", ""),
                'RT-DC data (*.rtdc)')
        if path:
            self.settings.setValue(
                "paths/open", str(pathlib.Path(path).parent))
            self.session_open(path)

    @QtCore.pyqtSlot()
    def on_action_quit(self, force=True):
        if force or self.session_close():
            QtCore.QCoreApplication.quit()

    @QtCore.pyqtSlot()
    def on_action_select_labels(self):
        """Let the user select a labeling group"""
        # get available labels
        groups = scores.get_available_label_groups()
        sel, ok = QtWidgets.QInputDialog.getItem(
            self,
            "Please select labeling group",
            "Please select a group for labeling",
            groups,
            editable=False)
        if ok:
            self.settings.setValue("labeling group", sel)
            QtWidgets.QMessageBox.information(
                self,
                "DCTag will now quit",
                "In order for the changes to take effect, "
                + "please restart DCTag."
            )
            if self.session_close():
                QtCore.QCoreApplication.quit()

    @QtCore.pyqtSlot()
    def on_action_software(self):
        libs = [dclab,
                h5py,
                numpy,
                ]
        sw_text = "DCTag {}\n\n".format(version)
        sw_text += "Python {}\n\n".format(sys.version)
        sw_text += "Modules:\n"
        for lib in libs:
            sw_text += "- {} {}\n".format(lib.__name__, lib.__version__)
        sw_text += "- PyQt5 {}\n".format(QtCore.QT_VERSION_STR)
        if hasattr(sys, 'frozen'):
            sw_text += "\nThis executable has been created using PyInstaller."
        QtWidgets.QMessageBox.information(self,
                                          "Software",
                                          sw_text)

    @QtCore.pyqtSlot()
    def on_tab_changed(self):
        curtab = self.tabWidget.currentWidget()
        curtab.update_session(self.session)
        self.set_title()

    def session_close(self):
        if not self.session:
            success = True
        else:
            try:
                self.session.flush()
            except session.DCTagSessionWriteError as e:
                QtWidgets.QMessageBox.warning(
                    self,
                    "Cannot close this session",
                    "For some reason, it is not possible to close the current "
                    + f"session '{self.session.path}'. Details:<br><br>"
                    + e.args[-1]
                )
                success = False
            else:
                self.session.close()
                self.session = None
                success = True
        return success

    @QtCore.pyqtSlot()
    def session_flush_statusbar(self):
        """Flush the session, writing all changes to the file

        If any errors occur, this is written to the status bar.
        """
        if self.session:
            date = time.strftime("%Y-%m-%d %H:%M:%S")
            try:
                self.session.flush()
            except BaseException as e:
                self.statusBar().showMessage(
                    f"{date} Saving failed with {e.__class__.__name__}: {e}")
                self.statusBar().setStyleSheet("color: red")
            else:
                self.statusBar().showMessage(f"{date} Session flushed.", 3000)
                self.statusBar().setStyleSheet("")

    def session_open(self, path_rtdc):
        """Load an .rtdc file into the user interface"""
        if self.session_close():
            user = self.settings.value("user/name", None)
            assert user
            # check whether we have a dctag-history log and if not,
            # ask the user whether to create a copy of the file.
            if session.is_dctag_session(path_rtdc):
                cont = True
            else:
                reply = QtWidgets.QMessageBox.question(
                    self,
                    "Claim this file?",
                    f"The file '{path_rtdc}' is not (yet) a DCTag session. "
                    + "Would you like to claim this file? If you select "
                    + "'Yes', this file will be tied to your username/alias. "
                    + "You may alternatively select "
                    + "'No' and open a copy of that file instead."
                )
                cont = reply == QtWidgets.QMessageBox.Yes
            if cont:
                try:
                    self.session = session.DCTagSession(path=path_rtdc,
                                                        user=user,
                                                        linked_features=[])
                except session.DCTagSessionWrongUserError as e:
                    reply_claim = QtWidgets.QMessageBox.question(
                        self,
                        f"Claim this file from {e.olduser}?",
                        f"This session is already claimed by {e.olduser}. "
                        f"Do you wish to force-claim this session anyway?"
                    )
                    if reply_claim == QtWidgets.QMessageBox.Yes:
                        self.session = session.DCTagSession(path=path_rtdc,
                                                            user=user,
                                                            linked_features=[],
                                                            override_user=True)
                    else:
                        # Don't do anything further here.
                        return
                # Go to session tab and update info
                self.tabWidget.setCurrentIndex(0)
                self.on_tab_changed()

    def set_title(self, task=None):
        if task is None:
            title = f"DCTag {version}"
        else:
            title = f"{task} [DCTag {version}]"
        self.setWindowTitle(title)


def excepthook(etype, value, trace):
    """
    Handler for all unhandled exceptions.

    :param `etype`: the exception type (`SyntaxError`,
        `ZeroDivisionError`, etc...);
    :type `etype`: `Exception`
    :param string `value`: the exception error message;
    :param string `trace`: the traceback header, if any (otherwise, it
        prints the standard Python header: ``Traceback (most recent
        call last)``.
    """
    vinfo = f"Unhandled exception in DCTag version {version}:\n"
    tmp = traceback.format_exception(etype, value, trace)
    exception = "".join([vinfo]+tmp)
    print(exception)

    errorbox = QtWidgets.QMessageBox()
    errorbox.setIcon(QtWidgets.QMessageBox.Critical)
    errorbox.addButton(QtWidgets.QPushButton('Close'),
                       QtWidgets.QMessageBox.YesRole)
    errorbox.addButton(QtWidgets.QPushButton(
        'Copy text && Close'), QtWidgets.QMessageBox.NoRole)
    errorbox.setText(exception)
    ret = errorbox.exec_()
    if ret == 1:
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(exception)


# Make Ctr+C close the app
signal.signal(signal.SIGINT, signal.SIG_DFL)
# Display exception hook in separate dialog instead of crashing
sys.excepthook = excepthook



import re

from dctag import scores

import pytest


def test_get_dctag_label_dict():
    blood = scores.get_dctag_label_dict(name="ml_scores_blood")
    assert blood["ml_score_r1f"]["label"] == "RBC singlet focused"


def test_get_dctag_label_dict_error_wrong_name():
    with pytest.raises(FileNotFoundError):
        scores.get_dctag_label_dict(name="peter")


@pytest.mark.parametrize("feat,label", [
    ["ml_score_r1f", "RBC singlet focused"],
    ["ml_score_66a", "ML score 66A"],  # from dclab
    ["userdef1", "User-defined 1"],  # from dclab
])
def test_get_feature_label(feat, label):
    assert scores.get_feature_label(feat) == label


@pytest.mark.parametrize("label,feat,shortcut", [
    ["ml_scores_blood", "ml_score_r1f", "R"],
    ["ml_scores_blood", "ml_score_r1u", "Ctrl+R"],
    ["ml_scores_blood", "ml_score_66a", "A"],  # default
    ["userdef", "userdef1", "1"],  # default
])
def test_get_feature_shortcut(label, feat, shortcut):
    assert scores.get_feature_shortcut(feat, label) == shortcut


def test_correct_ml_score_feat():
    blood = scores.get_dctag_label_dict(name="ml_scores_blood")
    pattern = re.compile("^ml_score_[a-z0-9]{3}$")
    for feat in blood.keys():
        assert pattern.match(feat), (f"Feature name '{feat}' does not match "
                                     "'ml_score_???'-pattern. Only lower case "
                                     "alphanumeric values allowed!")


def test_correct_ml_scores_format():
    blood = scores.get_dctag_label_dict(name="ml_scores_blood")
    for feat, feat_dict in blood.items():
        assert "label" in feat_dict, (f"Label entry missing for feature "
                                      f"'{feat}'!")
        assert "shortcut" in feat_dict, (f"Shortcut entry missing for feature "
                                         f"'{feat}'!")


def test_unique_score_labels():
    groups = scores.get_available_label_groups()
    for group in groups:
        blood = scores.get_dctag_label_dict(name=group)
        labels = [blood[ft]["label"] for ft in blood]
        assert len(labels) == len(set(labels))


def test_unique_score_shortcuts():
    blood = scores.get_dctag_label_dict(name="ml_scores_blood")
    shortcuts = [blood[ft]["shortcut"] for ft in blood]
    assert len(shortcuts) == len(set(shortcuts))



def main():
    import importlib.resources
    import sys

    from PyQt5.QtWidgets import QApplication

    app = QApplication(sys.argv)

    from PyQt5 import QtCore, QtGui
    from .gui import DCTag

    # Set Application Icon
    ref = importlib.resources.files("dctag.img") / "icon.png"
    with importlib.resources.as_file(ref) as icon_path:
        app.setWindowIcon(QtGui.QIcon(str(icon_path)))

    # Use dots as decimal separators
    QtCore.QLocale.setDefault(QtCore.QLocale(QtCore.QLocale.C))

    window = DCTag()  # noqa: F841
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



import io
import pathlib
import time
from unittest import mock

import h5py
import numpy as np
import pytest
from PyQt5 import QtCore, QtWidgets

import dctag
from dctag import session
from dctag.gui.main import DCTag

from .helper import get_clean_data_path


data_dir = pathlib.Path(__file__).parent / "data"


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 3000)
    pass
    # A test function will be run at this point
    yield
    # Code that will run after your test
    # restore dctag-tester for other tests
    QtCore.QCoreApplication.setOrganizationName("MPL")
    QtCore.QCoreApplication.setOrganizationDomain("dc-cosmos.org")
    QtCore.QCoreApplication.setApplicationName("dctag")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.setValue("user/name", "dctag-tester")
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 3000)


@pytest.mark.parametrize("with_delete", [True, False])
def test_action_backup(with_delete, qtbot, mw):
    # setup a nice session
    path = get_clean_data_path()
    # claim session
    with session.DCTagSession(path, "dctag-tester"):
        pass
    mw.on_action_open(path)
    # go through the tabs
    mw.tabWidget.setCurrentIndex(1)
    qtbot.mouseClick(mw.tab_binary.pushButton_start, QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_binary.pushButton_yes, QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_binary.pushButton_no, QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_binary.pushButton_yes, QtCore.Qt.LeftButton)

    if with_delete:
        path.unlink()

    export_path = path.with_name("export")

    # perform the export
    with mock.patch.object(QtWidgets.QMessageBox, "question",
                           return_value=QtWidgets.QMessageBox.Yes):
        with mock.patch.object(QtWidgets.QFileDialog, "getSaveFileName",
                               return_value=(str(export_path), None)):
            with mock.patch.object(mw, "on_action_quit"):
                mw.on_action_backup()

    export_path = export_path.with_suffix(".h5")

    assert export_path.exists()
    with h5py.File(export_path, "r") as h5:
        assert h5["ml_score_r1f"][0] == 1
        assert h5["ml_score_r1f"][1] == 0
        assert h5["ml_score_r1f"][2] == 1
        assert np.isnan(h5["ml_score_r1f"][3])
    # cleanup
    mw.session = None


def test_basic(qtbot, mw):
    """Run the program and exit"""
    time.sleep(.5)
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 3000)


def test_clear_session(mw):
    """Clearing the session should not cause any trouble"""
    path = get_clean_data_path()
    # claim session
    with session.DCTagSession(path, "dctag-tester"):
        pass
    # open session
    mw.on_action_open(path)
    # go through the tabs
    mw.tabWidget.setCurrentIndex(1)
    mw.tabWidget.setCurrentIndex(2)

    # Now clear the session
    mw.on_action_close()
    assert not mw.session
    # go through the tabs
    mw.tabWidget.setCurrentIndex(1)
    assert not mw.tab_binary.session
    assert not mw.tab_binary.widget_vis.session
    mw.tabWidget.setCurrentIndex(2)


def test_init_get_username(qtbot):
    # first reset the username
    # (undo what was done in conftest.py)
    QtCore.QCoreApplication.setOrganizationName("MPL")
    QtCore.QCoreApplication.setOrganizationDomain("dc-cosmos.org")
    QtCore.QCoreApplication.setApplicationName("dctag")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.remove("user/name")

    with mock.patch.object(QtWidgets.QInputDialog, "getText",
                           return_value=("peter", True)):
        mw = DCTag()
        mw.close()
        QtWidgets.QApplication.processEvents(
            QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 200)

    assert settings.value("user/name") == "peter"


def test_init_get_username_abort(qtbot):
    # first reset the username
    # (undo what was done in conftest.py)
    QtCore.QCoreApplication.setOrganizationName("MPL")
    QtCore.QCoreApplication.setOrganizationDomain("dc-cosmos.org")
    QtCore.QCoreApplication.setApplicationName("dctag")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.remove("user/name")

    with mock.patch.object(QtWidgets.QInputDialog, "getText",
                           return_value=("hans", False)):
        with pytest.raises(SystemExit):
            DCTag()
    QtWidgets.QApplication.processEvents(
        QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 200)

    assert settings.value("user/name") is None


def test_init_print_version(qtbot):
    mock_stdout = io.StringIO()
    mock_exit = mock.Mock()

    with mock.patch("sys.argv", ["--version"]):
        with mock.patch("sys.exit", mock_exit):
            with mock.patch('sys.stdout', mock_stdout):
                mw = DCTag()
                mw.close()
                QtWidgets.QApplication.processEvents(
                    QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 200)

    assert mock_exit.call_args.args[0] == 0
    assert mock_stdout.getvalue().strip() == dctag.__version__



import pyqtgraph as pg


class SimplePlotItem(pg.PlotItem):
    """Custom class for data visualization in dctag.

    Modifications include:
    - top and right axes
    """

    def __init__(self, *args, **kwargs):
        super(SimplePlotItem, self).__init__(*args, **kwargs)
        # show top and right axes
        for kax in ["top", "right", "left"]:
            self.showAxis(kax)
            ax = self.axes[kax]["item"]
            ax.setTicks([])
            ax.setLabel(None)
            ax.setStyle(tickTextOffset=0,
                        tickTextWidth=0,
                        tickTextHeight=0,
                        autoExpandTextSpace=False,
                        showValues=False,
                        )
        self.hideButtons()


class SimplePlotWidget(pg.PlotWidget):
    """Custom class for data visualization in dctag.

    Modifications include:
    - white background
    - those of SimplePlotItem
    """

    def __init__(self, parent=None, background='w', **kargs):
        plot_item = SimplePlotItem(**kargs)
        super(SimplePlotWidget, self).__init__(parent,
                                               background=background,
                                               plotItem=plot_item)
